@@ -60,6 +60,8 @@ class DataType:
             return self.is_list
         if other is dict:
             return self.is_struct
+        if other is object:
+            return self.is_object
         if isinstance(other, str):
             tester = 'is_' + other
             if hasattr(self, tester):
@@ -357,6 +359,11 @@ class DataType:
         '''
         return self.is_arrow and pa.types.is_struct(self.internal)
 
+    @property
+    def is_object(self):
+        '''Test if a NumPy dtype=object (avoid if possible)'''
+        return self.is_numpy and self.internal == object
+
     @property
     def is_encoded(self):
         '''Test if an (arrow) dictionary type (encoded data)
@@ -419,7 +426,10 @@ class DataType:
 
     def create_array(self, values):
         if self.is_arrow:
-            return pa.array(values, type=self.arrow)
+            if vaex.array_types.is_arrow_array(values):
+                return values
+            else:
+                return pa.array(values, type=self.arrow)
         else:
             return np.asarray(values, dtype=self.numpy)
 
